# This file is part of the Grid LSC User Environment (GLUE)
#
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

version = '2.0.0'

# git information
id = 'a1e5109f4893fd04cab4237827f62382ec61d50f'
date = '2019-02-20 22:43:17 +0000'
branch = 'master'
tag = 'glue-release-2.0.0'
author = 'Ryan Fisher <ryan.fisher@ligo.org>'
builder = 'Ryan Fisher <ryan.fisher@ligo.org>'
committer = 'Ryan Fisher <ryan.fisher@ligo.org>'
status = 'CLEAN: All modifications committed'
verbose_msg = """Branch: master
Tag: glue-release-2.0.0
Id: a1e5109f4893fd04cab4237827f62382ec61d50f

Builder: Ryan Fisher <ryan.fisher@ligo.org>
Build date: 2019-02-20 22:45:12 +0000
Repository status: CLEAN: All modifications committed"""

import warnings

class VersionMismatchError(ValueError):
    pass

def check_match(foreign_id, onmismatch="raise"):
    """
    If foreign_id != id, perform an action specified by the onmismatch
    kwarg. This can be useful for validating input files.

    onmismatch actions:
      "raise": raise a VersionMismatchError, stating both versions involved
      "warn": emit a warning, stating both versions involved
    """
    if onmismatch not in ("raise", "warn"):
        raise ValueError(onmismatch + " is an unrecognized value of onmismatch")
    if foreign_id == 'a1e5109f4893fd04cab4237827f62382ec61d50f':
        return
    msg = "Program id (a1e5109f4893fd04cab4237827f62382ec61d50f does not match given id (%s)." % foreign_id
    if onmismatch == "raise":
        raise VersionMismatchError(msg)

    # in the backtrace, show calling code
    warnings.warn(msg, UserWarning)





#
# ligolw_dq_active_cats: print state of defined cat vetos for a given
# time
#
# Copyright (C) 2009 Jameson Rollins and Josh Smith
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function
import sys
import os
from subprocess import *

############################################################

# usage
def usage():
	print("usage: LIGOLW_VETO_FILE=<file>", CMD, "<gpstime> [<duration>]")
	print("""
For specified time and category definer file, output flag states.  If
a duration is also specified, then the flag states are output for the
entire duration, starting at gpstime, in an OR manor, with preference
given to the flags in the order defined below.  The output is one line
per category-defined veto, of the form:

  <state> <cat> <ifo:flag>

The states are:

  + = active
  - = defined but inactive
  ! = undefined

The environment variable LIGOLW_VETO_FILE must contain the path or URL
of a category definer file.""")

# simple failure function
def failure(error='', errnum=1):
	sys.stderr.write(error)
	sys.exit(errnum)
	
# make category/start_time/end_time dictionaries from defined veto cat
# flags from a veto file.  returns three dictionaries with flag names
# as keys (<ifo>:<flag>) and categories, start times, and end times as
# values
def get_cats():
	cmd = 'ligolw_print -t veto_definer -d " " -c ifo -c name -c category -c start_time -c end_time %s' % (VETO_FILE)
	p1 = Popen(cmd, stdout=PIPE, shell=True)
	categories = {}
	start_times = {}
	end_times = {}
	for line in p1.communicate()[0].split('\n'):
		if line == '': continue
		info = line.split(' ')
		flag = info[0] + ':' + info[1]
		cat = info[2]
		start = float(info[3])
		end = float(info[4])
		if end == 0: end = float('inf')

		categories[flag] = cat
		start_times[flag] = start
		end_times[flag] = end

	if p1.wait() == 0:
		return categories, start_times, end_times
	else:
		failure("Could not retrieve info from veto definier file.\n")

# retrieve flags of type ('defined', 'active') for time
def get_flags(type, time, duration=0):
	# convert time to integer, since that's what dq_query accepts at the moment
	time = int(time)
	cmd1 = 'ligolw_dq_query --dmt-file --%s --end-pad %s %s' % (type, duration, time)
	cmd2 = 'ligolw_print -t segment_definer -c ifos -c name -d :'
	p1 = Popen(cmd1, stdout=PIPE, shell=True)
	p2 = Popen(cmd2, stdin=p1.stdout, stdout=PIPE, shell=True)
	flags = []
	[flags.append(line) for line in p2.communicate()[0].split('\n') if line != '']

	if p1.wait() == 0 and p2.wait() == 0:
		return flags
	else:
		failure("Could not retrieve %s flags.\n" % (type))
	
############################################################

CMD = os.path.basename(sys.argv.pop(0))

# single input argument is GPS time (or --help)
try:
	if sys.argv[0] in ['--help', '-h', '-help']:
		usage()
		sys.exit()
	else:
		time = float(sys.argv.pop(0))
except IndexError:
	failure("Must specify a GPS time. Type '--help' for more info.\n")
except ValueError:
	failure("GPS time must be a number. Type '--help' for more info.\n")

try:
	duration = int(sys.argv.pop(0))
except ValueError:
	failure("Duration must be an integer. Type '--help' for more info.\n")
except IndexError:
	duration = 0

# get the veto file from the environment
VETO_FILE = os.getenv("LIGOLW_VETO_FILE")
if VETO_FILE == None:
	failure("LIGOLW_VETO_FILE variable not defined.\n")

############################################################

# get categories, start and end times from definer file
categories, start_times, end_times = get_cats()

# retrieve active flags for time
actives = get_flags('active', time, duration)

# retrieve defined flags for time
defined = get_flags('defined', time, duration)

# process all vetos
for flag, category in categories.items():

	# check that the flag is valid for the time
	if time < start_times[flag]:
		continue
	if time >= end_times[flag]:
		continue

	# set the appropriate state indicator
	if flag in defined:
		if flag in actives:
			state = '+'
		else:
			state = '-'
	else:
		state = '!'

	# output the info
	print(state, category, flag)



#
#
# Copyright (C) 2009  Larne Pekowsky
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Takes two or more XML files containing segment tables, and calculates the
difference A - B - C - ... between either (1) the union of all specified segment definers
across files or (2) each individual segment definer between files
"""


from __future__ import print_function
from optparse import OptionParser

import sys
import os
import pwd

from glue.ligolw import utils
from glue.ligolw import ligolw
from glue.ligolw import lsctables

from glue.segmentdb import logic 

from glue.ligolw.utils import process

from glue.segmentdb.segmentdb_utils import add_to_segment_definer
from glue.segmentdb.segmentdb_utils import add_to_segment
from glue.segmentdb.segmentdb_utils import find_segments

from glue import git_version

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()

try:
    USER_NAME = os.getlogin()
except:
    USER_NAME = pwd.getpwuid(os.getuid())[0]


__author__  = "Larne Pekowsky <lppekows@physics.syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#
def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage   = "%prog [--segment | --segment-summary] [options] file1.xml file2.xml ...",
        description = "Takes two or more XML files containing segment tables, and calculates the\n" +
        "difference A - B - C - ... between either (1) the union of all specified segment definers\n" +
        "across files or (2) each individual segment definer between files\n"                      
	)
    
    parser.add_option("-o", "--output-file", metavar = "output_file", help = "Output file (default = stdout)")
    parser.add_option("-p", "--preserve",    action  = "store_true",  help = "If set, stores metadata from all input files in output")
    parser.add_option("-i", "--include-segments",    metavar = "include_segments",    help = "Segments to diff (ifo:name or ifo:name:version)")
    
    parser.add_option("-s", "--segment",           action = "store_true",   help = "Operate on the segment table")
    parser.add_option("-X", "--segment-summary",   action = "store_true",   help = "Operate on the segment_summary table")

    parser.add_option("-n", "--result-name", metavar = "result_name", default = "RESULT", help = "Name for result segment definer (default = RESULT)") 
    parser.add_option("-q", "--quiet",       action = "store_true",   help = "Set result status, but do not generate output")
    options, filenames = parser.parse_args()

    if not (options.segment or options.segment_summary) or (options.segment and options.segment_summary):
        print("Please specify one of --segment or  --segment-summary", file=sys.stderr)
        sys.exit(-1)

    return options, filenames



#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    options, filenames = parse_command_line()

    # Set up the document
    xmldoc = ligolw.Document()
    xmldoc.appendChild(ligolw.LIGO_LW())
    
    # Register ourselves
    proc_id = process.register_to_xmldoc(xmldoc, PROGRAM_NAME, options.__dict__, version=git_version.id).process_id

    # If we have no filenames, assume that we're reading from stdin
    # (ligolw_add uses None as a flag for stdin)
    if len(filenames) == 0:
        filenames = [None]

    # Also treat the special name 'stdin' as stdin
    filenames = [x != 'stdin' and x or None for x in filenames]

    if options.include_segments:
        xmldoc, total_time = logic.run_segment_operation(xmldoc, filenames, 
                                            options.include_segments,
                                            options.segment,
                                            logic.DIFF,
                                            options.result_name, options.preserve)
    else:
        xmldoc, total_time = logic.run_file_operation(xmldoc, filenames, options.segment, logic.DIFF, options.preserve)

    if not options.quiet:
        utils.write_filename(xmldoc, options.output_file)

    # Return 0 if there is no total time, 1 if there is
    # this is mostly useful for diff and intersect
    sys.exit(total_time and 1)




print("import: 'glue'")
import glue

print("import: 'glue.gpstime'")
import glue.gpstime

print("import: 'glue.iterutils'")
import glue.iterutils

print("import: 'glue.lal'")
import glue.lal

print("import: 'glue.ligolw'")
import glue.ligolw

print("import: 'glue.ligolw.ilwd'")
import glue.ligolw.ilwd

print("import: 'glue.ligolw.tokenizer'")
import glue.ligolw.tokenizer

print("import: 'glue.pipeline'")
import glue.pipeline

print("import: 'glue.text_progress_bar'")
import glue.text_progress_bar

print("import: 'glue.utils'")
import glue.utils

print("import: 'glue.workflow'")
import glue.workflow




# Copyright (C) 2014  Alex Nitz
#
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#
""" This module provides thin wrappers around Pegasus.DAX3 functionality that
provides additional abstraction and argument handling.
"""
import Pegasus.DAX3 as dax

class ProfileShortcuts(object):
    """ Container of common methods for setting pegasus profile information
    on Executables and nodes. This class expects to be inherited from
    and for a add_profile method to be implemented.
    """
    def set_memory(self, size):
        """ Set the amount of memory that is required in megabytes
        """
        self.add_profile('condor', 'request_memory', '%sM' % size)
         
    def set_storage(self, size):
        """ Set the amount of storage required in megabytes
        """
        self.add_profile('condor', 'request_disk', '%sM' % size)
        
    def set_num_cpus(self, number):
        self.add_profile('condor', 'request_cpus', number)
        
    def set_universe(self, universe):
        self.add_profile("condor", "universe", universe)
        
    def set_category(self, category):
        self.add_profile('dagman', 'category', category)
        
    def set_priority(self, priority):
        self.add_profile('dagman', 'priority', priority)
        
    def set_num_retries(self, number):
        self.add_profile("dagman", "retry", number)

class Executable(ProfileShortcuts):
    """ The workflow representation of an Executable 
    """
    id = 0
    def __init__(self, name, namespace=None, os='linux', 
                       arch='x86_64', installed=True, version=None):
        self.name = name
        self.logical_name = self.name + "_ID%s" % str(Executable.id)
        Executable.id += 1
        self.namespace = namespace
        self.version = version
        self._dax_executable = dax.Executable(self.logical_name, 
                   namespace=self.namespace, version=version, os=os, 
                   arch=arch, installed=installed) 
        self.in_workflow = False
        self.pfns = {}
        
    def add_pfn(self, url, site='local'):
        self._dax_executable.PFN(url, site)
        self.pfns[site] = url
        
    def get_pfn(self, site='local'):
        return self.pfns[site]
        
    def insert_into_dax(self, dax):
        dax.addExecutable(self._dax_executable)
        
    def add_profile(self, namespace, key, value):
        """ Add profile information to this executable
        """
        entry = dax.Profile(namespace, key, value)
        self._dax_executable.addProfile(entry)  
 
class Node(ProfileShortcuts):    
    def __init__(self, executable):
        self.in_workflow = False   
        self.executable=executable            
        self._inputs = []
        self._outputs = []        
        self._dax_node = dax.Job(name=executable.logical_name,
                                 version = executable.version, 
                                 namespace=executable.namespace)
        self._args = []
        self._options = []
        
    def add_arg(self, arg):
        """ Add an argument
        """
        if not isinstance(arg, File):
            arg = str(arg)
        
        self._args += [arg]
        
    def add_opt(self, opt, value=None):
        """ Add a option
        """
        if value:
            if not isinstance(value, File):
                value = str(value)
            self._options += [opt, value]
        else:
            self._options += [opt]
    
    #private functions to add input and output data sources/sinks        
    def _add_input(self, inp):
        """ Add as source of input data
        """
        self._inputs += [inp]
        inp._set_as_input_of(self)
      
    def _add_output(self, out):
        """ Add as destination of output data
        """
        self._outputs += [out]
        out.node = self
        out._set_as_output_of(self)
        
    # public functions to add options, arguments with or without data sources
    def add_input_opt(self, opt, inp):
        """ Add an option that determines an input
        """
        self.add_opt(opt, inp._dax_repr())
        self._add_input(inp)
        
    def add_output_opt(self, opt, out):
        """ Add an option that determines an output
        """
        self.add_opt(opt, out._dax_repr())
        self._add_output(out)
        
    def add_output_list_opt(self, opt, outputs):
        """ Add an option that determines a list of outputs
        """
        self.add_opt(opt)
        for out in outputs:
            self.add_opt(out)
            self._add_output(out)
            
    def add_input_list_opt(self, opt, inputs):
        """ Add an option that determines a list of inputs
        """
        self.add_opt(opt)
        for inp in inputs:
            self.add_opt(inp)
            self._add_input(inp)
        
    def add_input_arg(self, inp):
        """ Add an input as an argument
        """
        self.add_arg(inp._dax_repr())
        self._add_input(inp)
        
    def add_output_arg(self, out):
        """ Add an output as an argument
        """
        self.add_arg(out._dax_repr())
        self._add_output(out)
        
    def new_output_file_opt(self, opt, name):
        """ Add an option and return a new file handle
        """
        fil = File(name)
        self.add_output_opt(opt, fil)
        return fil

    # functions to describe properties of this node
    def add_profile(self, namespace, key, value):
        """ Add profile information to this node at the DAX level
        """
        entry = dax.Profile(namespace, key, value)
        self._dax_node.addProfile(entry)
        
    def _finalize(self):
        args = self._args + self._options
        self._dax_node.addArguments(*args)
        
class Workflow(object):
    """ 
    """
    def __init__(self, name='my_workflow'):
        self.name = name
        self._adag = dax.ADAG(name)
        
        self._inputs = []
        self._outputs = []
        self._executables = []
        
    def add_node(self, node):
        """ Add a node to this workflow
        
        This function adds nodes to the workflow. It also determines
        parent/child relations from the DataStorage inputs to this job. 
        
        Parameters
        ----------
        node : Node
            A node that should be exectuded as part of this workflow.
        """
        node._finalize()
        node.in_workflow = True
        self._adag.addJob(node._dax_node)
 
        # Determine the parent child relationships based on the inputs that
        # this node requires.
        for inp in node._inputs:
            if inp.node is not None and inp.node.in_workflow:
                parent = inp.node._dax_node
                child = node._dax_node
                dep = dax.Dependency(parent=parent, child=child)
                self._adag.addDependency(dep)    
                            
            elif inp.node is not None and not inp.node.in_workflow:
                raise ValueError('Parents of this node must be added to the '
                                 'workflow first.')   
                                  
            elif inp.node is None and inp.workflow_input is False:
                self._inputs += [inp]
                inp.workflow_input = True
            
        # Record the outputs that this node generates                                   
        self._outputs += node._outputs  
            
        # Record the executable that this node uses
        if not node.executable.in_workflow:
            node.executable.in_workflow = True
            self._executables += [node.executable]
        
        return self
        
    __iadd__ = add_node
      
    def save(self, filename):
        """ Write this workflow to DAX file 
        """
        f = open(filename, "w")
        self._adag.writeXML(f)

class DataStorage(object):
    """ A workflow representation of a place to store and read data from. 
    
    The abstract representation of a place to store and read data from. This 
    can include files, database, or remote connections. This object is
    used as a handle to pass between functions, and is used a way to logically
    represent the order operation on the physical data. 
    """
    def __init__(self, name):
        self.name = name      
        self.node = None
        self.workflow_input = False
        
    def _set_as_node_input(self):
        pass
        
    def _set_as_node_output(self):
        pass
        
    def _dax_repr(self):
        return self.name
        
class File(DataStorage, dax.File):
    """ The workflow representation of a physical file
    
    An object that represents a file from the perspective of setting up a 
    workflow. The file may or may not exist at the time of workflow generation.
    If it does, this is represented by containing a physical file name (PFN).
    A storage path is also available to indicate the desired final 
    destination of this file.
    """
    def __init__(self, name):
        DataStorage.__init__(self, name)
        dax.File.__init__(self, name)
        self.storage_path = None

    def _dax_repr(self):
        return self
        
    def _set_as_input_of(self, node):
        node._dax_node.uses(self, link=dax.Link.INPUT, register=False, 
                                                       transfer=True)          
    def _set_as_output_of(self, node):
        node._dax_node.uses(self, link=dax.Link.OUTPUT, register=False, 
                                                        transfer=True)                                                       
    def output_map_str(self):
        return '%s %s pool="%s"' % (self.name, self.storage_path, 'local') 
        
    def insert_into_dax(self, dax):
        dax.addFile(self)
    
class Database(DataStorage):
    pass



# Copyright (C) 2008-2010,2012-2015  Kipp Cannon
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Ask Kipp to document this!
"""


import itertools

from ligo import segments

from glue import git_version
from glue import iterutils
from ligo.segments import utils as segmentsUtils
from .. import ligolw
from .. import lsctables


__author__ = "Kipp Cannon <kipp.cannon@ligo.org>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                 Segment List
#
# =============================================================================
#


class LigolwSegmentList(object):
	"""
	A description of a LIGO Light-Weight XML segment list.  Instances
	of this class carry all the metadata associated with a LIGO Light-
	Weight XML segment list including its name, version number, a
	comment, and so on.

	LIGO Light-Weight XML segment lists are three-state objects.  A
	segment list can be on, off, or undefined.  Two separate sequences
	of segments are used for this:  the "valid" list defines the
	intervals when the state of the segment list is known, and the
	"active" list defines the intervals when the segment list is on.
	It is not an error for the active list to be on during times when
	the segment lists state is unknown, this code does not impose any
	policy in that regard, but it should be expected that consumers of
	the segment list will treat all times when the segment list's state
	is unknown the same way.

	Example:

	>>> from ligo.segments import *
	>>> segs = segmentlist([segment(0, 10), segment(20, 30)])
	>>> validity = segmentlist([segment(0, 10), segment(25, 100)])
	>>> x = LigolwSegmentList(active = segs, valid = validity, instruments = set(("H1",)), name = "test")
	>>> # x made copies of arguments
	>>> del segs[:]
	>>> segs
	[]
	>>> x.active
	[segment(0, 10), segment(20, 30)]
	>>> # some typical operations
	>>> x.active & x.valid	# known true
	[segment(0, 10), segment(25, 30)]
	>>> ~x.active & x.valid	# known false
	[segment(30, 100)]
	>>> x.active & ~x.valid	# not an error for this to be non-null
	[segment(20, 25)]
	>>> # make a copy
	>>> y = LigolwSegmentList(x)
	>>> del y.active[:]
	>>> y.active
	[]
	>>> x.active
	[segment(0, 10), segment(20, 30)]

	The arithmetic operators on this class implement Kleene's strong
	ternary logic, taking "true" to be (active & valid), "false" to be
	(~active & valid), and "unknown" to be ~valid.

	Example:

	>>> from ligo.segments import *
	>>> segs = segmentlist([segment(0, 10), segment(20, 30)])
	>>> validity = segmentlist([segment(0, 35)])
	>>> x = LigolwSegmentList(active = segs, valid = validity, instruments = set(("H1",)), name = "test")
	>>> segs = segmentlist([segment(40, 50), segment(60, 70)])
	>>> validity = segmentlist([segment(35, 100)])
	>>> y = LigolwSegmentList(active = segs, valid = validity, instruments = set(("H1",)), name = "test")
	>>> (x | y).active
	[segment(0, 10), segment(20, 30), segment(40, 50), segment(60, 70)]
	>>> (x | y).valid
	[segment(0, 10), segment(20, 30), segment(40, 50), segment(60, 70)]
	>>> (x & y).active
	[]
	>>> (x & y).valid
	[segment(10, 20), segment(30, 40), segment(50, 60), segment(70, 100)]
	>>> (~x).active
	[segment(10, 20), segment(30, 35)]
	>>> (~x).valid
	[segment(0, 35)]

	With ternary logic the three basic Boolean operations AND, OR, and
	NOT, do not form a complete set of operations.  That is, there
	exist algebraic functions that cannot be implemented using
	combinations of these three operators alone.  One additional
	operator is required to construct a complete basis of logic
	operations, and we provide one:  .isfalse().  This operation
	inverts intervals of known state, and maps intervals of unknown
	state to false.

	>>> x.isfalse().active
	[segment(10, 20), segment(30, 35)]
	>>> x.isfalse().valid
	[segment(-infinity, infinity)]

	Unfortunately, one example of a function that cannot be constructed
	from the three basic Boolean operators is perhaps the most common
	operation we wish to perform with our tri-state segment lists.
	Often we wish to construct a tri-state list from two tri-state
	lists such that the final list's interval of validity is the union
	of the intervals of validity of the two source lists, and the state
	of the final list in that interval is the union the states of the
	source lists in that interval.  For example if from one source we
	know the state of some process spanning some time, and from another
	source we know the state of the same process spanning some other
	time, taken together we know the state of that process over the
	union of those times.  This function is given by

	>>> z = ~(x.isfalse() | y.isfalse() | (x & ~x & y & ~y))
	>>> z.active
	[segment(0, 10), segment(20, 30), segment(40, 50), segment(60, 70)]
	>>> z.valid
	[segment(0, 100)]

	Because this is inconvenient to type, slow, and not readable, a
	special in-place arithmetic operation named .update() is provided
	to implement this operation.

	>>> z = LigolwSegmentList(x).update(y)
	>>> z.active
	[segment(0, 10), segment(20, 30), segment(40, 50), segment(60, 70)]
	>>> z.valid
	[segment(0, 100)]

	The .update() method is not exactly equivalent to the operation
	above.  The .update() method demands that the two input lists'
	states be identical where their intervals of validity intersect.
	"""
	#
	# the columns the segment_definer, segment_summary and segment
	# tables need to have
	#

	segment_def_columns = (u"process_id", u"segment_def_id", u"ifos", u"name", u"version", u"comment")
	segment_sum_columns = (u"process_id", u"segment_sum_id", u"start_time", u"start_time_ns", u"end_time", u"end_time_ns", u"segment_def_id", u"comment")
	segment_columns = (u"process_id", u"segment_id", u"start_time", u"start_time_ns", u"end_time", u"end_time_ns", u"segment_def_id")

	def __init__(self, active = (), valid = (), instruments = (), name = None, version = None, comment = None):
		"""
		Initialize a new LigolwSegmentList instance.  active and
		valid are sequences that will be cast to
		segments.segmentlist objects.  They can be generator
		expressions.  The "active" sequence is what is usually
		thought of as the segment list, the "valid" sequence
		identifies the intervals of time for which the segment
		list's state is defined.
		"""
		# if we've only been passed an argument for active, see if
		# it's an object with the same attributes as ourselves and
		# if so initialize ourself as a copy of it.
		if not valid and not instruments and name is None and version is None and comment is None:
			try:
				self.valid = segments.segmentlist(active.valid)
				self.active = segments.segmentlist(active.active)
				self.instruments = set(active.instruments)
				self.name = active.name
				self.version = active.version
				self.comment = active.comment
				return
			except AttributeError:
				pass
		# we had more than one argument or it didn't have the
		# correct attributes, so do a normal initialization.  make
		# copies of mutable objects to avoid confusion
		self.valid = segments.segmentlist(valid)
		self.active = segments.segmentlist(active)
		self.instruments = set(instruments)
		self.name = name
		self.version = version
		self.comment = comment

	def sort(self, *args):
		"""
		Sort the internal segment lists.  The optional args are
		passed to the .sort() method of the segment lists.  This
		can be used to control the sort order by providing an
		alternate comparison function.  The default is to sort by
		start time with ties broken by end time.
		"""
		self.valid.sort(*args)
		self.active.sort(*args)

	def coalesce(self):
		"""
		Coalesce the internal segment lists.  Returns self.
		"""
		self.valid.coalesce()
		self.active.coalesce()
		return self

	def __ior__(self, other):
		"""
		If either is true the result is true, if both are false the
		result is false, otherwise the result is unknown.
		"""
		if self.instruments != other.instruments:
			raise ValueError("incompatible metadata")
		# times when one or the other is known to be true
		self.active &= self.valid
		self.active |= other.active & other.valid
		# times when result is defined
		self.valid = (self.valid & other.valid) | self.active
		return self

	def __iand__(self, other):
		"""
		If either is false the result is false, if both are true
		the result is true, otherwise the result is unknown.
		"""
		if self.instruments != other.instruments:
			raise ValueError("incompatible metadata")
		# times when one or the other is known to be false
		false = (self.valid & ~self.active) | (other.valid & ~other.active)
		# times when both are known to be true
		self.active &= self.valid
		self.active &= other.active & other.valid
		# times when result is defined
		self.valid = false | self.active
		return self

	def __or__(self, other):
		"""
		If either is true the result is true, if both are false the
		result is false, otherwise the result is unknown.
		"""
		result = type(self)(self)
		result |= other
		return result

	def __and__(self, other):
		"""
		If either is false the result is false, if both are true
		the result is true, otherwise the result is unknown.
		"""
		result = type(self)(self)
		result &= other
		return result

	def __invert__(self):
		"""
		If unknown the result is unknown, otherwise the state is
		inverted.
		"""
		result = type(self)(self)
		result.active = ~result.active & result.valid
		return result

	def isfalse(self):
		"""
		If unknown the result is false, otherwise the state is
		inverted.
		"""
		result = type(self)(self)
		result.active = ~result.active & result.valid
		result.valid = segments.segmentlist([segments.segment(-segments.infinity(), +segments.infinity())])
		return result

	def update(self, other):
		if self.instruments != other.instruments:
			raise ValueError("incompatible metadata")
		if (self.valid & other.valid).intersects(self.active ^ other.active):
			raise ValueError("result over-determined")
		self.active &= self.valid
		self.active |= other.active & other.valid
		self.valid |= other.valid
		return self


#
# =============================================================================
#
#                                 Library API
#
# =============================================================================
#


class LigolwSegments(set):
	"""
	An interface shim between code that makes use of segments in
	ligo.segments form, and LIGO Light-Weight XML I/O code.

	This class is "attached" to an XML document object, at which time
	it parses and extracts the segment lists from the document, and
	clears the document's segment tables (preventing a second
	LigolwSegments object from being meaningfully attached to the same
	document).  When the application is finished manipulating the
	segment lists, they can be inserted into the XML document at which
	time the contents of the LigolwSegments object are cleared
	(preventing any further manipulations).

	This class is a subclass of the Python set builtin.  Each element
	of the set is a LigolwSegmentList instance describing one of the
	segment lists in the original XML document.

	This class may be used as a context manager to automate the
	replacement of segments back into the XML document, including in
	the event of an untrapped exception.  When used as a context
	manager, the process parameter of the .__init__() method is not
	optional.

	Example:

	>>> import sys
	>>> from ligo.segments import *
	>>> from lal import LIGOTimeGPS
	>>> from glue.ligolw import ligolw, lsctables
	>>> xmldoc = ligolw.Document()
	>>> xmldoc.appendChild(ligolw.LIGO_LW())	# doctest: +ELLIPSIS
	<glue.ligolw.ligolw.LIGO_LW object at ...>
	>>> process = lsctables.Process(process_id = lsctables.ProcessTable.get_next_id())
	>>> with LigolwSegments(xmldoc, process) as xmlsegments:
	...	h1segs = segmentlist([segment(LIGOTimeGPS(0), LIGOTimeGPS(10))])
	...	xmlsegments.insert_from_segmentlistdict({"H1": h1segs}, "test")
	...	l1segs = h1segs.shift(5)
	...	xmlsegments.add(LigolwSegmentList(active = l1segs, valid = segmentlist([segment(-infinity(), infinity())]), instruments = set(["L1"]), name = "test"))
	>>> xmldoc.write(sys.stdout)		# doctest: +NORMALIZE_WHITESPACE
	<?xml version='1.0' encoding='utf-8'?>
	<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
	<LIGO_LW>
		<Table Name="segment_definer:table">
			<Column Name="segment_definer:comment" Type="lstring"/>
			<Column Name="segment_definer:ifos" Type="lstring"/>
			<Column Name="segment_definer:name" Type="lstring"/>
			<Column Name="segment_definer:process_id" Type="ilwd:char"/>
			<Column Name="segment_definer:segment_def_id" Type="ilwd:char"/>
			<Column Name="segment_definer:version" Type="int_4s"/>
			<Stream Delimiter="," Name="segment_definer:table" Type="Local">
				,"H1","test","process:process_id:0","segment_definer:segment_def_id:0",,
				,"L1","test","process:process_id:0","segment_definer:segment_def_id:1",,
			</Stream>
		</Table>
		<Table Name="segment_summary:table">
			<Column Name="segment_summary:comment" Type="lstring"/>
			<Column Name="segment_summary:end_time" Type="int_4s"/>
			<Column Name="segment_summary:end_time_ns" Type="int_4s"/>
			<Column Name="segment_summary:process_id" Type="ilwd:char"/>
			<Column Name="segment_summary:segment_def_id" Type="ilwd:char"/>
			<Column Name="segment_summary:segment_sum_id" Type="ilwd:char"/>
			<Column Name="segment_summary:start_time" Type="int_4s"/>
			<Column Name="segment_summary:start_time_ns" Type="int_4s"/>
			<Stream Delimiter="," Name="segment_summary:table" Type="Local">
				,2147483647,4294967295,"process:process_id:0","segment_definer:segment_def_id:1","segment_summary:segment_sum_id:0",4294967295,4294967295
			</Stream>
		</Table>
		<Table Name="segment:table">
			<Column Name="segment:end_time" Type="int_4s"/>
			<Column Name="segment:end_time_ns" Type="int_4s"/>
			<Column Name="segment:process_id" Type="ilwd:char"/>
			<Column Name="segment:segment_def_id" Type="ilwd:char"/>
			<Column Name="segment:segment_id" Type="ilwd:char"/>
			<Column Name="segment:start_time" Type="int_4s"/>
			<Column Name="segment:start_time_ns" Type="int_4s"/>
			<Stream Delimiter="," Name="segment:table" Type="Local">
				10,0,"process:process_id:0","segment_definer:segment_def_id:0","segment:segment_id:0",0,0,
				15,0,"process:process_id:0","segment_definer:segment_def_id:1","segment:segment_id:1",5,0
			</Stream>
		</Table>
	</LIGO_LW>
	>>> xmlsegments = LigolwSegments(xmldoc)
	>>> assert xmlsegments.get_by_name("test") == {'H1': [segment(LIGOTimeGPS(0, 0), LIGOTimeGPS(10, 0))], 'L1': [segment(LIGOTimeGPS(5, 0), LIGOTimeGPS(15, 0))]}
	>>> xmlsegments.get_by_name("wrong name")
	Traceback (most recent call last):
		...
	KeyError: "no segmentlists named 'wrong name'"

	NOTE:  the process of extracting and re-inserting the contents of
	the segment tables will, in general, randomize the IDs assigned to
	the rows of these tables.  If there are references to segment,
	segment_summary, or segment_definer row IDs in other tables in the
	document, those references will be broken by this process.
	"""
	def __init__(self, xmldoc, process = None):
		#
		# Find tables
		#

		try:
			self.segment_def_table = lsctables.SegmentDefTable.get_table(xmldoc)
		except ValueError:
			self.segment_def_table = xmldoc.childNodes[0].appendChild(lsctables.New(lsctables.SegmentDefTable, LigolwSegmentList.segment_def_columns))

		try:
			self.segment_sum_table = lsctables.SegmentSumTable.get_table(xmldoc)
		except ValueError:
			self.segment_sum_table = xmldoc.childNodes[0].appendChild(lsctables.New(lsctables.SegmentSumTable, LigolwSegmentList.segment_sum_columns))

		try:
			self.segment_table = lsctables.SegmentTable.get_table(xmldoc)
		except ValueError:
			self.segment_table = xmldoc.childNodes[0].appendChild(lsctables.New(lsctables.SegmentTable, LigolwSegmentList.segment_columns))

		#
		# Transform segment tables into a collection of
		# LigolwSegmentList objects for more convenient
		# manipulation
		#

		# construct empty LigolwSegmentList objects, one for each
		# entry in the segment_definer table, indexed by
		# segment_definer id
		segment_lists = dict((row.segment_def_id, LigolwSegmentList(instruments = row.instruments, name = row.name, version = row.version, comment = row.comment)) for row in self.segment_def_table)
		if len(segment_lists) != len(self.segment_def_table):
			raise ValueError("duplicate segment_def_id in segment_definer table")
		del self.segment_def_table[:]

		# populate LigolwSegmentList objects from segment table and
		# segment_summary table
		for row in self.segment_sum_table:
			try:
				segment_lists[row.segment_def_id].valid.append(row.segment)
			except KeyError as e:
				raise ValueError("invalid segment_def_id in segment_summary table: %s" % e)
		del self.segment_sum_table[:]
		for row in self.segment_table:
			try:
				segment_lists[row.segment_def_id].active.append(row.segment)
			except KeyError as e:
				raise ValueError("invalid segment_def_id in segment table: %s" % e)
		del self.segment_table[:]

		#
		# transcribe LigolwSegmentList objects into self.  segment
		# definer IDs are now meaningless
		#

		self.update(segment_lists.values())

		#
		# reset ID generators
		#

		self.segment_def_table.set_next_id(type(self.segment_def_table.next_id)(0))
		self.segment_table.set_next_id(type(self.segment_table.next_id)(0))
		self.segment_sum_table.set_next_id(type(self.segment_sum_table.next_id)(0))

		#
		# Save process row for later
		#

		self.process = process

		#
		# Done
		#


	def insert_from_segwizard(self, fileobj, instruments, name, version = None, comment = None):
		"""
		Parse the contents of the file object fileobj as a
		segwizard-format segment list, and insert the result as a
		new list of "active" segments into this LigolwSegments
		object.  A new entry will be created in the segment_definer
		table for the segment list, and instruments, name and
		comment are used to populate the entry's metadata.  Note
		that the "valid" segments are left empty, nominally
		indicating that there are no periods of validity.  Returns
		the newly created LigolwSegmentList object.
		"""
		ligolw_segment_list = LigolwSegmentList(active = segmentsUtils.fromsegwizard(fileobj, coltype = lsctables.LIGOTimeGPS), instruments = instruments, name = name, version = version, comment = comment)
		self.add(ligolw_segment_list)
		return ligolw_segment_list


	def insert_from_segmentlistdict(self, seglists, name, version = None, comment = None):
		"""
		Insert the segments from the segmentlistdict object
		seglists as a new list of "active" segments into this
		LigolwSegments object.  The dictionary's keys are assumed
		to provide the instrument name for each segment list.  A
		new entry will be created in the segment_definer table for
		the segment lists, and the dictionary's keys, the name, and
		comment will be used to populate the entry's metadata.
		"""
		for instrument, segments in seglists.items():
			self.add(LigolwSegmentList(active = segments, instruments = set([instrument]), name = name, version = version, comment = comment))


	def coalesce(self):
		"""
		Coalesce the segment lists.  Returns self.
		"""
		for ligolw_segment_list in self:
			ligolw_segment_list.coalesce()
		return self


	def sort(self, *args):
		"""
		Sort the segment lists.  The optional args are passed to
		the .sort() methods of the segment lists.  This can be used
		to control the sort order by providing an alternate
		comparison function (the default is to sort all lists by
		segment start time with ties broken by end time).
		"""
		for ligolw_segment_list in self:
			ligolw_segment_list.sort(*args)


	def optimize(self):
		"""
		Identifies segment lists that differ only in their
		instruments --- they have the same valid and active
		segments, the same name, version and the same comment ---
		and then deletes all but one of them, leaving just a single
		list having the union of the instruments.
		"""
		self.sort()
		segment_lists = dict(enumerate(self))
		for target, source in [(idx_a, idx_b) for (idx_a, seglist_a), (idx_b, seglist_b) in itertools.combinations(segment_lists.items(), 2) if seglist_a.valid == seglist_b.valid and seglist_a.active == seglist_b.active and seglist_a.name == seglist_b.name and seglist_a.version == seglist_b.version and seglist_a.comment == seglist_b.comment]:
			try:
				source = segment_lists.pop(source)
			except KeyError:
				continue
			segment_lists[target].instruments |= source.instruments
		self.clear()
		self.update(segment_lists.values())


	def get_by_name(self, name, clip_to_valid = False):
		"""
		Retrieve the active segmentlists whose name equals name.
		The result is a segmentlistdict indexed by instrument.  All
		segmentlist objects within it will be copies of the
		contents of this object, modifications will not affect the
		contents of this object.  If clip_to_valid is True then the
		segmentlists will be intersected with their respective
		intervals of validity, otherwise they will be the verbatim
		active segments.

		NOTE:  the intersection operation required by clip_to_valid
		will yield undefined results unless the active and valid
		segmentlist objects are coalesced.
		"""
		result = segments.segmentlistdict()
		for seglist in self:
			if seglist.name != name:
				continue
			segs = seglist.active
			if clip_to_valid:
				# do not use in-place intersection
				segs = segs & seglist.valid
			for instrument in seglist.instruments:
				if instrument in result:
					raise ValueError("multiple '%s' segmentlists for instrument '%s'" % (name, instrument))
				# make copy so that instruments do not
				# share a single list, so that subsequent
				# arithmetic operations do not corrupt the
				# wrong instrument's segments
				result[instrument] = segments.segmentlist(segs)
		if not result:
			raise KeyError("no segmentlists named '%s'" % name)
		return result


	def finalize(self, process_row = None):
		"""
		Restore the LigolwSegmentList objects to the XML tables in
		preparation for output.  All segments from all segment
		lists are inserted into the tables in time order, but this
		is NOT behaviour external applications should rely on.
		This is done simply in the belief that it might assist in
		constructing well balanced indexed databases from the
		resulting files.  If that proves not to be the case, or for
		some reason this behaviour proves inconvenient to preserve,
		then it might be discontinued without notice.  You've been
		warned.
		"""
		if process_row is not None:
			process_id = process_row.process_id
		elif self.process is not None:
			process_id = self.process.process_id
		else:
			raise ValueError("must supply a process row to .__init__()")

		#
		# ensure ID generators are synchronized with table contents
		#

		self.segment_def_table.sync_next_id()
		self.segment_table.sync_next_id()
		self.segment_sum_table.sync_next_id()

		#
		# put all segment lists in time order
		#

		self.sort()

		#
		# generator function to convert segments into row objects,
		# each paired with the table to which the row is to be
		# appended
		#

		def row_generator(segs, target_table, process_id, segment_def_id):
			id_column = target_table.next_id.column_name
			for seg in segs:
				row = target_table.RowType()
				row.segment = seg
				row.process_id = process_id
				row.segment_def_id = segment_def_id
				if isinstance(row, lsctables.SegmentSum):
					row.comment = None
				yield row, target_table, id_column

		#
		# populate the segment_definer table from the list of
		# LigolwSegmentList objects and construct a matching list
		# of table row generators.  empty ourselves to prevent this
		# process from being repeated
		#

		row_generators = []
		for ligolw_segment_list in sorted(self, key = lambda l: (l.name, sorted(l.instruments), l.version)):
			self.remove(ligolw_segment_list)
			segment_def_row = self.segment_def_table.RowType()
			segment_def_row.process_id = process_id
			segment_def_row.segment_def_id = self.segment_def_table.get_next_id()
			segment_def_row.instruments = ligolw_segment_list.instruments
			segment_def_row.name = ligolw_segment_list.name
			segment_def_row.version = ligolw_segment_list.version
			segment_def_row.comment = ligolw_segment_list.comment
			self.segment_def_table.append(segment_def_row)

			row_generators.append(row_generator(ligolw_segment_list.valid, self.segment_sum_table, process_id, segment_def_row.segment_def_id))
			row_generators.append(row_generator(ligolw_segment_list.active, self.segment_table, process_id, segment_def_row.segment_def_id))

		#
		# populate segment and segment_summary tables by pulling
		# rows from the generators in time order
		#

		for row, target_table, id_column in iterutils.inorder(*row_generators):
			setattr(row, id_column, target_table.get_next_id())
			target_table.append(row)


	def __enter__(self):
		if self.process is None:
			raise ValueError("must supply a process row to .__init__()")
		return self


	def __exit__(self, *args):
		self.finalize()
		return False


#
# =============================================================================
#
#                                  Utilities
#
# =============================================================================
#


@lsctables.use_in
class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
	"""
	Minimal content handler suitable for loading documents containg
	segment tables.
	"""
	pass


def has_segment_tables(xmldoc, name = None):
	"""
	Return True if the document contains a complete set of segment
	tables.  Returns False otherwise.  If name is given and not None
	then the return value is True only if the document's segment
	tables, if present, contain a segment list by that name.
	"""
	try:
		names = lsctables.SegmentDefTable.get_table(xmldoc).getColumnByName("name")
		lsctables.SegmentTable.get_table(xmldoc)
		lsctables.SegmentSumTable.get_table(xmldoc)
	except (ValueError, KeyError):
		return False
	return name is None or name in names


def segmenttable_get_by_name(xmldoc, name):
	"""
	Retrieve the segmentlists whose name equals name.  The result is a
	segmentlistdict indexed by instrument.

	The output of this function is not coalesced, each segmentlist
	contains the segments as found in the segment table.

	NOTE:  this is a light-weight version of the .get_by_name() method
	of the LigolwSegments class intended for use when the full
	machinery of that class is not required.  Considerably less
	document validation and error checking is performed by this
	version.  Consider using that method instead if your application
	will be interfacing with the document via that class anyway.
	"""
	#
	# find required tables
	#

	def_table = lsctables.SegmentDefTable.get_table(xmldoc)
	seg_table = lsctables.SegmentTable.get_table(xmldoc)

	#
	# segment_def_id --> instrument names mapping but only for
	# segment_definer entries bearing the requested name
	#

	instrument_index = dict((row.segment_def_id, row.instruments) for row in def_table if row.name == name)

	#
	# populate result segmentlistdict object from segment_def_map table
	# and index
	#

	instruments = set(instrument for instruments in instrument_index.values() for instrument in instruments)
	result = segments.segmentlistdict((instrument, segments.segmentlist()) for instrument in instruments)

	for row in seg_table:
		if row.segment_def_id in instrument_index:
			seg = row.segment
			for instrument in instrument_index[row.segment_def_id]:
				result[instrument].append(seg)

	#
	# done
	#

	return result



#
# Copyright (C) 2009  Larne Pekowsky, based on glitch-page.sh by Duncan
# Brown
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from __future__ import print_function
from optparse import OptionParser
import sqlite3
import sys
import os
import glob
import tempfile

from ligo.segments import segmentlist, segment

from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils
from glue.ligolw.utils import ligolw_sqlite
from glue.ligolw import dbtables

from glue.segmentdb import query_engine
from glue.segmentdb import segmentdb_utils

from glue import gpstime

import glob
import time
import datetime
import StringIO

#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#

def parse_command_line():
    """
    Parse the command line, return an options object
    """

    def require(flag, message):
        if not flag:
            print(message, file=sys.stdout)
            sys.exit(1)

    parser = OptionParser(
        version     = "%prog CVS $Header$",
        usage       = "%prog --trigger-dir dir --segments url --html-file file --ifo ifo --timestamp-file [other options]",
        description = "Updates or creates the html with recnet glitch summary information"
	)

    parser.add_option("-t", "--trigger-dir",   metavar="dir",  help = "Location of XML files containing sngl_inspiral tables")
    parser.add_option("-e", "--trigger-db",    metavar="database",  help = "Location of an already-created sqlite database")
    parser.add_option("-s", "--segments",      metavar="url",  help = "URL to contact for DQ flags (ldbd: or file:)")
    parser.add_option("-f", "--html-file",     metavar="file", help = "Location of html file to write")
    parser.add_option("-i", "--ifo",           metavar="ifo",  help = "IFO")
    parser.add_option("-g", "--timestamp-file", metavar="file", help = "Location of file storing last run time")
    parser.add_option("-m", "--min-glitch-snr", metavar="snr",  help = "Minimum SNR to be considered a glitch (default=15.0)",  default="15.0")
    parser.add_option("-k", "--known-count",    metavar="known_count", help = "Max. number of triggers with DQ flags to print (default=10)", default="10")
    parser.add_option("-u", "--unknown-count",  metavar="unknown_count", help = "Max. number of triggers without DQ flags to print (default=10)", default="10")
    parser.add_option("-b", "--gps-start-time",  metavar="gps_start_time", help = "Provide an explicit start time, rather than using the timestamp file")
    parser.add_option("-n", "--gps-end-time",  metavar="gps_end_time", help = "Provide an explicit end time, rather than using now")

    parser.add_option("-a", "--omega-conf",       metavar="omega_conf",      help = "Omega configuration file")
    parser.add_option("-c", "--omega-frame-dir",  metavar="omega_frame_dir", help = "Top level directory of frame files for omega")
    parser.add_option("-d", "--omega-out-dir",    metavar="omega_out_dir",   help = "Output directory for omega results")
    parser.add_option("-r", "--statistic",        metavar="statistic",       help = "Ranking statistic, snr|newsnr, default=snr", default="snr")

    options, others = parser.parse_args()

    # Make sure we have all required parameters
    require(options.trigger_dir or options.trigger_db, "Missing required argument [--trigger-dir | --trigger-db]")
    require(options.ifo, "Missing required argument --ifo")
    require(options.html_file, "Missing required argument --html-file")
    require(options.timestamp_file or (options.gps_start_time and options.gps_end_time), "Missing required argument [--timestamp-file | --gps-start-time --gps-end-time]")

    return options



def sigfigs(num):
    if num == 0.0:
        return num

    if num < 10:
        div = 1.0

        while num < 10:
            num *= 10.0
            div *= 10.0

        return '%.1f' % (int(num) / div)

    mul = 1

    while num > 100:
        num /= 10.0
        mul *= 10

    return str(int(num) * mul)


def generate_html(outf, triggers, colors, options):
    print('<p>', file=outf)
    print('<table border=1>', file=outf)

    if options.statistic == 'newsnr':
        print('  <tr bgcolor="#9999ff"><th>ifo</th><th>end time UTC</th>', end=' ', file=outf)
        print('<th>end_time</th><th>end_time_ns</th>', end=' ', file=outf)
        print('<th>SNR</th><th>&chi;<sup>2</sup></th><th>New SNR</th><th>eff_distance</th>', file=outf)
        print("<th>M<sub>chirp</sub></th>", end=' ', file=outf)
        print("<th>M<sub>total</sub></th>", end=' ', file=outf)
        print("<th>&Omega; scan</th><th>DQ flags</th></tr>", file=outf)
    else:
        print('  <tr bgcolor="#9999ff"><th>ifo</th><th>end time UTC</th><th>end_time</th><th>end_time_ns</th><th>SNR</th><th>eff_distance</th>', file=outf)
        print("<th>M<sub>chirp</sub></th>", end=' ', file=outf)
        print("<th>M<sub>total</sub></th>", end=' ', file=outf)
        print("<th>&chi;<sup>2</sup></th>", end=' ', file=outf)
        print("<th>Reduced cont. &chi;<sup>2</sup></th>", end=' ', file=outf)
        print("<th>&Omega; scan</th><th>DQ flags</th></tr>", file=outf)

    for count, trig in enumerate(triggers):
        print('  <tr valign="top" bgcolor="%s">' % colors[count % 2], file=outf)

        print('    <td>%s</td>' % (trig[0]), file=outf)

        end_time_utc = os.popen('lalapps_tconvert %d' % trig[1]).next().strip()
        print('    <td>%s</td>' % (end_time_utc), file=outf)

        for res in trig[1:-1]:
            if str(res).find('.') > 0:
                res = sigfigs(float(res))

            print('    <td>%s</td>' % res, file=outf)
        
        print('    <td><a href="omega/%s.%s/index.html">&Omega; scan</a></td>' % (trig[1], trig[2]), file=outf)

        print('    <td>', file=outf)
        for name, value in trig[-1].items():
            if name.find('Light') != -1 or name.find('Up') != -1 or name.find('Calibrated') != -1:
                print('      <b>%s</b><br>' % name, file=outf)
            else:
                print('      %s %s %s<br>' % (name, value[0], value[2]), file=outf)
        print('    </td>', file=outf)

        print('  </tr>', file=outf)
    print('</table>', file=outf)

    print(file=outf)
    print(file=outf)




def setup_files(dir_name, db_name, gps_start_time, gps_end_time):
    if db_name:
        temp_db    = None
        connection = sqlite3.connect(db_name)
    else:
        # Filter out the ones that are outside our time range
        xml_files  = segmentdb_utils.get_all_files_in_range(dir_name, gps_start_time, gps_end_time)

        handle, temp_db = tempfile.mkstemp(suffix='.sqlite')
        os.close(handle)

        target     = dbtables.get_connection_filename(temp_db, None, True, False)
        connection = ligolw_sqlite.setup(target)

        ligolw_sqlite.insert_from_urls(connection, xml_files)

    return temp_db, connection
    

def get_times_with_flags(gps_start_time, gps_end_time):
    # Find times where there were no DQ flags, rather than checking the time of each trigger
    segment_connection = segmentdb_utils.setup_database(os.environ['S6_SEGMENT_SERVER'])
    engine             = query_engine.LdbdQueryEngine(segment_connection)

    sql  = "SELECT start_time, end_time FROM segment "
    sql += " WHERE NOT (%s > segment.end_time OR segment.start_time > %s) " % (gps_start_time, gps_end_time)
    sql += " AND segment.segment_def_id IN (SELECT DISTINCT segment_definer.segment_def_id FROM segment_definer, segment_summary "
    sql += "                                WHERE segment_definer.ifos = '%s' " % options.ifo
    sql += "                                AND segment_definer.segment_def_id = segment_summary.segment_def_id "
    sql += "                                AND segment_summary.segment_def_cdb = segment_definer.creator_db "
    sql += "                                AND segment_definer.name NOT IN ('DMT-SCIENCE', 'DMT-LIGHT', 'DMT-UP', 'DMT-CALIBRATED', 'DMT-INJECTION') "
    sql += "                                AND NOT (%s > segment.end_time OR segment.start_time > %s)) " % (gps_start_time, gps_end_time)

    times_with_flags = segmentlist([segment(row[0], row[1]) for row in engine.query(sql)])
    times_with_flags.coalesce()

    return times_with_flags


def newsnr(snr,chisq,chisq_dof):
    index  = 6.0
    rchisq = chisq/(2 * chisq_dof - 2)
    nhigh  = 2.

    if rchisq > 1.:
        return snr / ((1. + rchisq**(index/nhigh))/2)**(1./index)
    else:
        return snr


def oldsnr(snr,chisq,chisq_dof):
    return snr

def calc_mchirp(mass1,mass2):
    return pow(mass1 * mass2, 3.0/5.0) / pow(mass1 + mass2, 1.0/5.0)

def get_triggers_in_times(options, gps_start_time, gps_end_time, valid_times, count):
    if not abs(valid_times):
        return []

    # Note, the S5 version of this script had the condition
    #    search = 'FindChirpSPtwoPN' 
    # The triggers from MBTA don't set this, alough it could if desirable.
    # 
    # When hitting DB2 we could use the following query:
    #
    #  SELECT sngl_inspiral.ifo, 
    #         sngl_inspiral.end_time,
    #         sngl_inspiral.end_time_ns,
    #         sngl_inspiral.snr, 
    #         sngl_inspiral.eff_distance,
    #         sngl_inspiral.f_final, 
    #         sngl_inspiral.ttotal
    #  FROM sngl_inspiral 
    #  WHERE (sngl_inspiral.end_time, sngl_inspiral.snr) IN 
    #         (select end_time, MAX(snr) AS snr 
    #          FROM sngl_inspiral
    #          WHERE end_time >= ? AND
    #                end_time <  ? AND
    #                ifo = ? AND
    #                snr >= 15.0 
    #          GROUP BY end_time
    #          ORDER BY snr desc)
    #  AND sngl_inspiral.ifo = ? 
    #  ORDER BY snr DESC""",  (start_time, end_time, ifo, ifo) )
    #
    # But sqlite doesn't allow this, since:
    #   SQL error: only a single result allowed for a SELECT that is part of an expression
    #
    # So we have to be a little trickier...
    #

    connection.create_function("mchirp", 2, calc_mchirp)

    if options.statistic == 'newsnr':
        connection.create_function("newsnr", 3, newsnr)
        fields = """sngl_inspiral.ifo, 
             sngl_inspiral.end_time,
             sngl_inspiral.end_time_ns,
             sngl_inspiral.snr,
             sngl_inspiral.chisq,
             newsnr(sngl_inspiral.snr,sngl_inspiral.chisq,sngl_inspiral.chisq_dof), 
             sngl_inspiral.eff_distance,
             mchirp(sngl_inspiral.mass1,sngl_inspiral.mass2),
             sngl_inspiral.mass1 + sngl_inspiral.mass2
        """
    else:
        connection.create_function("newsnr", 3, oldsnr)
        fields = """sngl_inspiral.ifo, 
             sngl_inspiral.end_time,
             sngl_inspiral.end_time_ns,
             sngl_inspiral.snr,
             sngl_inspiral.eff_distance,
             sngl_inspiral.chisq,
             sngl_inspiral.mass1,
             sngl_inspiral.mass2,
             sngl_inspiral.bank_chisq / (2.0 * sngl_inspiral.bank_chisq_dof),
             sngl_inspiral.cont_chisq / sngl_inspiral.cont_chisq_dof
        """

    #
    # The join above was meant to ensure that we only pick
    # the loudest trigger from each second, so we don't
    # fill the table with triggers from the same glitch.
    # Bu since we only ever look at 16-second clustered
    # triggers anyway, we don't need that.
    #
    rows = connection.cursor().execute("SELECT " + fields + """
      FROM sngl_inspiral 
      WHERE sngl_inspiral.end_time >= ? AND
            sngl_inspiral.end_time <  ? AND
            sngl_inspiral.ifo = ? AND
            newsnr(sngl_inspiral.snr, sngl_inspiral.chisq, sngl_inspiral.chisq_dof) >= ?
      ORDER BY newsnr(sngl_inspiral.snr,sngl_inspiral.chisq,sngl_inspiral.chisq_dof) DESC""",
      (gps_start_time, gps_end_time, options.ifo, float(options.min_glitch_snr)) )

    ret = []
    i   = 0

    # These variables names are incorrect if we're using new snr: snr appears before new_snr and
    # cont_chisq is eleimated.  But we only really need the time fields, and there's the same 
    # number of values.
    for ifo, end_time, end_time_ns, snr, chisq, new_snr, eff_dist, mchirp, mtotal in rows:
        trig_time = end_time
        if end_time_ns >= 500000000:
            trig_time += 1
    
        if trig_time not in valid_times:
            continue

        # Find the flags on at this time.  We need this even for "unknown" times, since
        # we need to check for DMT-SCIENCE etc
        flags = {}

        pipe  = os.popen('ligolw_dq_query --segment=%s --include-segments %s --in-segments-only --report %d' % (options.segments, ifo, end_time))
        for line in pipe:
            flag, beforet, timet, aftert = [x for x in line.split() if x != '']
    
            ifo, name, version = flag.split(':')
            flags[name] = (beforet, timet, aftert)

        pipe.close()
    
        ifo_status = ''

        for flag_name in ['Light','Up','Calibrated','Science','Injection']:
            flag = 'DMT-' + flag_name.upper()

            if flag in flags:
                ifo_status += flag_name + ','
                del flags[flag]

        if len(ifo_status) > 0:
            ifo_status        = ifo_status[:-1]
            flags[ifo_status] = True

        ret.append((ifo, end_time, end_time_ns, snr, chisq, new_snr, eff_dist, mchirp, mtotal, flags))


        i += 1

        if i > count:
            break

    return ret


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    options = parse_command_line()    

    # Do we have everything we need for omega scans?
    omega_ok = options.omega_conf and options.omega_frame_dir and options.omega_out_dir

    if not omega_ok:
        print("Warning: Missing omega options, omega scans will not be run", file=sys.stderr)


    kcount = int(options.known_count)
    ucount = int(options.unknown_count)

    # Run up to now or the user-provided time
    if options.gps_end_time:
        gps_end_time = int(options.gps_end_time)
    else:
        gps_end_time = gpstime.GpsSecondsFromPyUTC(time.time())

    # Find the last time we ran (in principle we could get this from the time
    # stamp on the HTML file, but we should allow for the possibility that 
    # someone might hand-edit that file)
    if options.gps_start_time:
        gps_start_time = int(options.gps_start_time)
    elif os.path.exists(options.timestamp_file):
        f = open(options.timestamp_file)
        gps_start_time = int(next(f))
        f.close()
    else:
        # Looks like we've never run before, start a day ago
        gps_start_time = gps_end_time - (60 * 60 * 24)


    # Load the relevant trigger XML files into a sqlite DB and
    # get a connection
    temp_db, connection = setup_files(options.trigger_dir, options.trigger_db, gps_start_time, gps_end_time)

    # Did we find any triggers?  If not there won't even be a sngl_inspiral
    # table, so the normal query will fail

    have_triggers = connection.cursor().execute("SELECT COUNT(*) FROM sqlite_master WHERE name = 'sngl_inspiral'").fetchone()[0]

    if have_triggers: 
        times_with_flags    = get_times_with_flags(gps_start_time, gps_end_time)
        times_without_flags = segmentlist([segment(gps_start_time,gps_end_time)]) - times_with_flags

        known_trigs   = get_triggers_in_times(options, gps_start_time, gps_end_time, times_with_flags, kcount)
        unknown_trigs = get_triggers_in_times(options, gps_start_time, gps_end_time, times_without_flags, ucount)
    else:
        # Use an empty array so the rest of the code will flow through and
        # update the page accordingly
        known_trigs    = []
        unknown_trigs  = []

    # ligolw_sicluster --cluster-window ${clusterwindow} --sort-descending-snr ${xmlfile}

    # Do the omega scans if requested
    if omega_ok:
        for trig in (known_trigs + unknown_trigs):
            ifo         = trig[0]
            end_time    = trig[1]
            end_time_ns = trig[2]

            # The frames may live in a directory with subdirectories by time.  If the given directory
            # contains a '%d' populate it with the first 4 digits of the time.
            full_path = options.omega_frame_dir

            if full_path.find('%d') != -1:
                full_path = full_path.replace('%d', str(end_time / 100000))

            if full_path.find('%s') != -1:
                full_path = full_path.replace('%s', ifo[0])

            conf = options.omega_conf.replace('%s',ifo[0])

            os.system("nohup condor_run ~omega/opt/omega/bin/wpipeline scan -r -c %s -f %s -o %s/%d.%d  %d.%d < /dev/null &>/dev/null &" % ( conf, full_path, options.omega_out_dir, end_time, end_time_ns, end_time, end_time_ns ))


    # Convert to html, prepend to the file.
    out_tmp = StringIO.StringIO()
   
    # Could these be done in Python?
    start_time_str = os.popen('lalapps_tconvert %d' % gps_start_time).next().strip()
    end_time_str   = os.popen('lalapps_tconvert %d' % gps_end_time).next().strip()

    print("<h3>%s through %s</h3>" % (start_time_str, end_time_str), file=out_tmp)

    print('<h3>Glitches without associated DQ flag</h3>', file=out_tmp)
    generate_html(out_tmp, unknown_trigs, ['#ffdddd', '#ffcccc'], options)
    

    print('<h3>Glitches with associated DQ flag</h3>', file=out_tmp)
    generate_html(out_tmp, known_trigs, ['#ddffdd', '#ccffcc'], options)

    print("<hr>", file=out_tmp)

    if os.path.exists(options.html_file):
        in_tmp = open(options.html_file)
        for l in in_tmp:
            print(l, end=' ', file=out_tmp)
        in_tmp.close()

    out_html = open(options.html_file,'w')
    print(out_tmp.getvalue(), file=out_html)
    out_html.close()

    # All done, if we weren't given an ending time update the timestamp
    if not options.gps_end_time:
        f = open(options.timestamp_file, 'w')
        print(gps_end_time, file=f)
        f.close()

    # If we created our own database, clean up after ourselves.
    if temp_db:
        os.remove(temp_db)






#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


from six.moves import zip
description = \
"""
The executable front-end of print_tables.py. This program prints tables in
xml documents in specified formats -- currently wiki or html.
"""
import sys, os
from optparse import OptionParser

from glue import git_version
from glue.ligolw import ligolw
from glue.ligolw import table
from glue.ligolw import utils
from glue.ligolw.utils import ligolw_add
from glue.ligolw.utils import print_tables


__prog__ = "ligolw_print_tables"
__author__ = "Collin Capano <cdcapano@physics.syr.edu>"


# =============================================================================
#
#                                   Set Options
#
# =============================================================================


def parse_command_line():
    """
    Parse the command line, return options and check for consistency among the
    options.
    """
    parser = OptionParser(
        version = git_version.verbose_msg,
        usage   = "%prog [--output-format] [options] file1.xml file2.xml ...",
        description = description
        )
    
    parser.add_option("-o", "--output", action = "store", type = "string",
        default = sys.stdout,
        help =
            "Save converted tables to a file. If no output specified, result " +
            "will be printed to stdout."
            )
    parser.add_option("-f", "--output-format", action = "store", type = "string",
        default = None, metavar = "wiki OR html",
        help =
            "Requried. Format to convert xml tables into. Current choices are 'wiki' or 'html'."
            )
    parser.add_option("-m", "--merge-docs", action = "store_true",
        default = False,
        help =
            "Merge all xml files into a single set of tables. WARNING: this will cause " +
            "all ilwd:char (e.g., coinc_event_ids) to be re-mapped in the output."
            )
    parser.add_option("-t", "--table", action = "append", type = "string",
        default = [],
        help =
            "Only convert the specified table (default is to convert all the tables). " +
            "To specify multiple tables, give the argument multiple times. " +
            "All the tables must exist in the xml file(s)."
            )
    parser.add_option("-c", "--column", action = "append", type = "string",
        default = [],
        help =
            "Only print the given column (default is to print all columns). " +
            "This applies to all tables with a column having the given name(s)." +
            "To specify a column in a specific table, use table_name:column_name. " +
            "To specify multiple columns, give the argument multiple times (order given " +
            "will be the order printed). "
            )
    parser.add_option("-s","--rspan-column", action = "append", type = "string",
        default = [],
        help =
            "Allow column to row-span; i.e., consecutive rows that have the same value " +
            "in the given column will be combined into one cell that spans those rows. " +
            "Default is to not row span any column. To specify multiple columns, give the " +
            "argument multiple times."
            )
    parser.add_option("-b", "--break-column", action = "append", type = "string",
        default = [],
        help =
            "Whenever a value between consecutive rows in the given column change, " +
            "no columns will be allowed to row-span across that row."
            )
    parser.add_option("-r", "--round-floats", action = "store", type = "int",
        default = None,
        help =
            "Round floating point numbers in the tables to the specified " +
            "number of decimal places."
            )
    parser.add_option("-l", "--format-links", action = "store_true",
        default = False,
        help =
            'Convert any entry with <a href=" in it to a hyperlink for the given ' +
            'format. Note: if output format is html, this is a no-op.'
            )
    parser.add_option("", "--no-table-names", action = "store_true",
        default = False,
        help =
            "Turn off printing table names."
            )

    (options, filenames) = parser.parse_args()

    if not options.output_format:
        raise ValueError("--output-format is a required option")

    return options, filenames


#
# =============================================================================
#
#                                   Main
#
# =============================================================================
#


options, filenames = parse_command_line()

# Also treat the special name 'stdin' as stdin
filenames = [x != 'stdin' and x or None for x in filenames]

# get the xmldocs
if options.merge_docs:
    xmldocs = [ ligolw_add.ligolw_add( ligolw.Document(), filenames, non_lsc_tables_ok = True ) ]
    titles = [ None ]
else:
    xmldocs = [ utils.load_filename( filename, gz = (filename or "stdin").endswith(".gz") ) for filename in filenames ]
    titles = filenames

# open the output
if options.output != sys.stdout:
    options.output = open(options.output, 'w')

if options.round_floats is None:
    round_floats = False
else:
    round_floats = True

print_table_names = not options.no_table_names

# convert the files
for xmldoc, title in zip(xmldocs, titles):
    print_tables.print_tables(xmldoc, options.output, options.output_format, tableList = options.table,
        columnList = options.column, round_floats = round_floats, decimal_places = options.round_floats,
        format_links = options.format_links, row_span_columns = options.rspan_column, rspan_break_columns = options.break_column,
        title = title, print_table_names = print_table_names)

# close and exit
options.output.close()

sys.exit(0)



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Text-mode progress bars
"""
from __future__ import division, print_function, unicode_literals
from six.moves import range

import locale
import math
import os
import struct
import sys
import collections

__all__ = ('ProgressBar', 'ProgressBarTheme')


# From http://stackoverflow.com/questions/566746
def getTerminalSize():
    """
    returns (lines:int, cols:int)
    """

    def ioctl_GWINSZ(fd):
        # These two imports are only present on POSIX systems, so they must be
        # guarded by a try block.
        import fcntl
        import termios
        return struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
    # try stdin, stdout, stderr
    for fd in (0, 1, 2):
        try:
            return ioctl_GWINSZ(fd)
        except:
            pass
    # try os.ctermid()
    try:
        fd = os.open(os.ctermid(), os.O_RDONLY)
        try:
            return ioctl_GWINSZ(fd)
        finally:
            os.close(fd)
    except:
        pass
    # try environment variables
    try:
        return tuple(int(os.getenv(var)) for var in ("LINES", "COLUMNS"))
    except:
        pass
    # i give up. return default.
    return (25, 80)


_ProgressBarTheme = collections.namedtuple(
    '_ProgressBarTheme', 'sequence twiddle_sequence left_border right_border')


class ProgressBarTheme(_ProgressBarTheme):

    def is_compatible_with_encoding(self, coding):
        if not coding:
            coding = locale.getpreferredencoding()
        try:
            for string in self:
                string.encode(coding)
        except UnicodeEncodeError:
            return False
        else:
            return True

    def is_compatible_with_stream(self, stream):
        return self.is_compatible_with_encoding(stream.encoding)


default_unicode_theme = ProgressBarTheme(
    ' ▏▎▍▌▋▊▉█', '       ▏▎▍▌▋▊▉██▉▊▋▌▍▎▏ ', '▐', '▌')


default_ascii_theme = ProgressBarTheme(
    ' .:!|', ' ..', ' |', u'| ')


class ProgressBar:
    """Display a text progress bar.

    A final line feed is printed when the ProgressBar is garbage collected.
    Explicitly deleting the object can force a line feed when desired.  As
    an alternative, using the ProgressBar as a context manager will ensure
    a final line feed is printed when the code block within which the
    ProgressBar is being used exits.

    Example:

    >>> with ProgressBar(max=3) as pb:
    ...    pb.update(1)
    ...    pb.update(2)
    ...    pb.update(3)
    ...
    """

    def __init__(
            self, text='Working', max=1, value=0, textwidth=24, fid=None,
            theme=None):
        if fid is None:
            self.fid = sys.stderr
        if hasattr(self.fid, 'fileno'):
            self.isatty = os.isatty(self.fid.fileno())
        else:
            self.isatty = False
        if theme is None:
            if self.isatty and default_unicode_theme.is_compatible_with_stream(
                    self.fid) and 'xterm' in os.environ.get('TERM', ''):
                theme = default_unicode_theme
            else:
                theme = default_ascii_theme
        self.text = text
        self.max = max
        self.value = value
        self.textwidth = textwidth
        self.sequence = ('',) + tuple(theme.sequence)
        self.twiddle_sequence = tuple(
            theme.twiddle_sequence[-i:] + theme.twiddle_sequence[:-i]
            for i in range(len(theme.twiddle_sequence)))
        self.left_border = theme.left_border
        self.right_border = theme.right_border
        self.twiddle = 0
        self.linefed = False

    def iterate(self, iterable, format="%s", print_every=1):
        """Use as a target of a for-loop to issue a progress update for every
        iteration. For example:

        progress = ProgressBar()
        for text in progress.iterate(["foo", "bar", "bat"]):
            ...
        """

        # If iterable has a definite length, then set the maximum value of the
        # progress bar. Else, set the maximum value to -1 so that the progress
        # bar displays indeterminate progress (scrolling dots).
        try:
            length = len(iterable)
        except TypeError:
            self.max = -1
        else:
            self.max = length

        # Iterate over the input, updating the progress bar for each element.
        for i, item in enumerate(iterable):
            yield item
            if i % print_every == 0:
                self.update(i + 1, format % item)

    def show(self):
        """Redraw the text progress bar."""
        # Be silent if writing to a tty
        if not self.isatty:
            return

        if len(self.text) > self.textwidth:
            label = self.text[:self.textwidth]
        else:
            label = self.text.rjust(self.textwidth)

        terminalSize = getTerminalSize()
        if terminalSize is None:
            terminalSize = 80
        else:
            terminalSize = terminalSize[1]

        barWidth = terminalSize - self.textwidth - len(self.left_border) \
            - len(self.right_border) - 7

        if self.value is None or self.value < 0:
            pattern = self.twiddle_sequence[
                self.twiddle % len(self.twiddle_sequence)]
            self.twiddle += 1
            barSymbols = (pattern * int(math.ceil(barWidth / len(self.twiddle_sequence))))[0:barWidth]
            progressFractionText = '      '
        else:
            progressFraction = max(0.0, min(1.0, float(self.value) / self.max))

            fMinor, iMajor = math.modf(progressFraction * barWidth)
            iMajor = int(iMajor)
            iMinor = int(math.ceil(fMinor * (len(self.sequence) - 1)))
            iMajorMinor = int(math.ceil(progressFraction * barWidth))

            barSymbols = (
                (self.sequence[-1] * iMajor) +
                self.sequence[iMinor] +
                (self.sequence[1] * (barWidth - iMajorMinor)))
            progressFractionText = ('%.1f%%' % (100*progressFraction)).rjust(6)

        print(
            '\r\x1B[1m', label, '\x1B[0m', self.left_border, '\x1B[36m',
            barSymbols, '\x1B[0m', self.right_border, progressFractionText,
            sep='', end='', file=self.fid)
        self.fid.flush()
        self.linefed = False

    def update(self, value=None, text=None):
        """Redraw the progress bar, optionally changing the value and text
        and return the (possibly new) value.  For I/O performance, the
        progress bar might not be written to the terminal if the text does
        not change and the value changes by too little.  Use .show() to
        force a redraw."""
        redraw = False
        if text is not None:
            redraw = text != self.text
            self.text = text
        if value is not None:
            redraw |= self.max == 0 or round(value/(0.0003*self.max)) != \
                round(self.value/(0.0003*self.max))
            self.value = value
        if redraw:
            self.show()
        return self.value

    def increment(self, delta=1, text=None):
        """Redraw the progress bar, incrementing the value by delta
        (default=1) and optionally changing the text.  Returns the
        ProgressBar's new value.  See also .update()."""
        return self.update(value=min(self.max, self.value + delta), text=text)

    def linefeed(self):
        # Be silent if writing to a tty
        if not self.isatty:
            return
        if not self.linefed:
            print(file=self.fid)
            self.fid.flush()
            self.linefed = True

    def __enter__(self):
        self.show()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self.linefeed()
        except:
            pass

    def __del__(self):
        self.linefeed()


def demo():
    """Demonstrate progress bar."""
    from time import sleep
    maxProgress = 1000
    with ProgressBar(max=maxProgress) as progressbar:
        for i in range(-100, maxProgress):
            sleep(0.01)
            progressbar.update(i + 1)
    progressbar2 = ProgressBar(max=maxProgress)
    for s in progressbar2.iterate(list(range(maxProgress))):
        sleep(0.01)
    for s in progressbar2.iterate(
            list(range(maxProgress)), format='iteration %d'):
        sleep(0.01)


if __name__ == '__main__':
    demo()
